/*
 * linux/arch/mips/vr41xx/irq.c
 *
 * Code to handle VR41XX IRQs plus some generic interrupt stuff.
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1994, 1995, 1996, 1997 Ralf Baechle
 * Copyright (C) 1999 Bradley D. LaRonde and Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/malloc.h>
#include <linux/random.h>

#include <asm/bitops.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/vr41xx.h>

// The following three are not used here, but needed elsewhere
unsigned int local_bh_count[NR_CPUS];
unsigned int local_irq_count[NR_CPUS];
unsigned long spurious_count = 0;

unsigned short irq_mask_probe[5];

#define BIT_MASK(bit) ( 1 << (bit) )
#define MAKE_CP0_STATUS_IRQ_MASK(irq) ( BIT_MASK( (irq) + 8))

static inline void mask_irq(unsigned int irq)
{
	if (irq < 8) {
		// it's a cpu interrupt
		unsigned short newstatus = read_32bit_cp0_register(CP0_STATUS);
		newstatus &= ~((unsigned short)1 << (irq + 8)); 
		set_cp0_status(ST0_IM, newstatus);
	} else {
		if (irq < 40) {
			// it's an ICU interrupt
			if (irq < 24) {
				*VR41XX_MSYSINT1REG &= ~((unsigned short)1 << (irq - 8));
			} else {
				*VR41XX_MSYSINT2REG &= ~((unsigned short)1 << (irq - 24));
			}
		} else {
			// it's a GPIO interrupt
			if (irq < 56) {
				*VR41XX_MGIUINTLREG &= ~((unsigned short)1 << (irq - 40));
			} else {
				*VR41XX_MGIUINTHREG &= ~((unsigned short)1 << (irq - 56));
			}
		}
	}
}

static inline void unmask_irq(unsigned int irq)
{
	if (irq < 8) {
		// it's a cpu interrupt
		unsigned short newstatus = read_32bit_cp0_register(CP0_STATUS);
		newstatus |= ((unsigned short)1 << (irq + 8)); 
		set_cp0_status(ST0_IM, newstatus);
	} else {
		if (irq < 40) {
			// it's an ICU interrupt
			if (irq < 24) {
				*VR41XX_MSYSINT1REG |= (unsigned short)1 << (irq - 8);
			} else {
				*VR41XX_MSYSINT2REG |= (unsigned short)1 << (irq - 24);
			}
		} else {
			// it's a GPIO interrupt
			if (irq < 56) {
				*VR41XX_MGIUINTLREG |= (unsigned short)1 << (irq - 40);
			} else {
				*VR41XX_MGIUINTHREG |= (unsigned short)1 << (irq - 56);
			}
		}
	}
}

void disable_irq(unsigned int irq_nr)
{
    unsigned long flags;

    save_and_cli(flags);
    mask_irq(irq_nr);
    restore_flags(flags);
}

void enable_irq(unsigned int irq_nr)
{
    unsigned long flags;

    save_and_cli(flags);
    unmask_irq(irq_nr);
    restore_flags(flags);
}

/*
 * Pointers to the low-level handlers: first the general ones, then the
 * fast ones, then the bad ones.
 */


static struct irqaction* irq_action[VR41XX_IRQ_MAX + 1] =
{
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

int get_irq_list(char *buf)
{
	// make a human-readable list of irq actions
	// used by /proc/interrupts

	int i, len = 0;
	struct irqaction *action;

	for (i = 0; i <= VR41XX_IRQ_MAX; i++) {
		action = irq_action[i];
		if (!action)
			continue;
		len += sprintf(buf + len, "%2d: %8d %c %s",
			i, kstat.irqs[0][i],
			(action->flags & SA_INTERRUPT) ? '+' : ' ',
			action->name);
		for (action = action->next; action; action = action->next) {
			len += sprintf(buf + len, ",%s %s",
				(action->flags & SA_INTERRUPT) ? " +" : "",
				action->name);
		}
		len += sprintf(buf + len, "\n");
   	 }
	return len;
}

atomic_t __mips_bh_counter;

asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
	// This handles all IRQ's that have been installed.
	// It is called from int-handler.S.
	// Actions without SA_INTERRUPT run with interrupts enabled
	// and use the full signal-handling return.
	// Actions with SA_INTERRUPT run with interrupts disabled.

	// Previous comment said: All relatively slow IRQ's should use [the SA_INTERRUPT] format:
	// notably the keyboard/timer routines.

	struct irqaction *action;
	int cpu = smp_processor_id();

	// enter the irq
	hardirq_enter(cpu);

	// count this irq on this cpu
	kstat.irqs[cpu][irq]++;

	// don't interrupt on this same irq again until we're finished
	// also, it gets left masked if there is no action (see below)
	mask_irq(irq);

	// figure a pointer to the irq action list list
	action = irq_action[irq];

	// is there an action list?
	if (action != 0) {
		unsigned long flags = 0;

		// should this handler run with interrupts enabled?
		if (!(action->flags & SA_INTERRUPT))
			__sti();

		// for each action in the list
		do {
			// handle it
			action->handler(irq, action->dev_id, regs);

			flags |= action->flags;
			action = action->next;
		} while (action);

		// were any of the actions SA_SAMPLE_RANDOM?
		if (flags & SA_SAMPLE_RANDOM)
			add_interrupt_randomness(irq);

		__cli();

		// let this irq cause interrupts again
		unmask_irq(irq);
	}

	// exit the irq
	hardirq_exit(cpu);

	// Unmasking and bottom half handling is done for us
	// currently by ret_from_irq in entry.S.
}

int add_irq_action(int irq, struct irqaction *new)
{
	// put the specified action (new) at the end of the list of actions for this irq

	int shared = 0;
	struct irqaction *old, **p;
	unsigned long flags;

	// pointer to the first action in the list for this irq
	p = irq_action + irq;

	// already a list there?
	if ((old = *p) != NULL) {
		// Is this action asking to share when the first one didn't? 
		if (!(old->flags & new->flags & SA_SHIRQ))
		    return -EBUSY;

		// Isn't this action the same type as the first action in the list?
		if ((old->flags ^ new->flags) & SA_INTERRUPT)
		    return -EBUSY;

		// Find the last action in this list 
		do {
		    p = &old->next;
		    old = *p;
		} while (old);

		// there's more than one action in this list
		shared = 1;
	}

	// is this a SA_SAMPLE_RANDOM action?
	if (new->flags & SA_SAMPLE_RANDOM)
		// initialize the randomness
		rand_initialize_irq(irq);

	save_and_cli(flags);

	// add action to the end of the list
	*p = new;

	// Isn't it a shared irq?
	if (!shared) {
		unmask_irq(irq);
	}

	restore_flags(flags);

	return 0;
}

int request_irq(unsigned int irq,
	void (*handler)(int, void *, struct pt_regs *),
	unsigned long irqflags,
	const char *devname,
	void *dev_id)
{
	int retval;
	struct irqaction *action;

	// some assertiveness
	if (irq > VR41XX_IRQ_MAX)
		return -EINVAL;

	if (!handler)
		return -EINVAL;

	// allocate a new action struct
	action = (struct irqaction *) kmalloc(sizeof(struct irqaction), GFP_KERNEL);
	if (!action)
		return -ENOMEM;

	// initialize it
	action->handler = handler;
	action->flags = irqflags;
	action->mask = 0;
	action->name = devname;
	action->next = NULL;
	action->dev_id = dev_id;

	// add it to the list of actions for this irq
	retval = add_irq_action(irq, action);
	if (retval)
		// so much for that
		kfree(action);

	return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
    struct irqaction *action, **p;
    unsigned long flags;

    if (irq > VR41XX_IRQ_MAX) {
	printk("Error - trying to free invalid IRQ %d\n", irq);
	return;
    }

    for (p = irq + irq_action; (action = *p) != NULL; p = &action->next) {
	if (action->dev_id != dev_id)
	    continue;

	/* Found it - now free it */
	save_and_cli(flags);
	*p = action->next;
	if (!irq[irq_action])
	    mask_irq(irq);
	restore_flags(flags);
	kfree(action);
	return;
    }
}

unsigned long probe_irq_on(void)
{
	int i;
	unsigned long delay;

	for (i = 0; i < 5; i++)
		irq_mask_probe[i] = 0;

	// Note the lower 8 bits of irq_mask_probe[0] are not used
	
	for (i = VR41XX_IRQ_MAX; i > 0; i--) {
		if (!irq_action[i]) {
			enable_irq(i);
			irq_mask_probe[(i + 8)/16] |= (unsigned short)1 << ((i + 8) & 15);
		}
	}

	// Wait for spurious interrupts to mask themselves out again...
	for (delay = jiffies + HZ/10; time_before(jiffies, delay); )
		barrier();	// about 100ms of delay

	irq_mask_probe[0] &= read_32bit_cp0_register(CP0_STATUS);
	irq_mask_probe[1] &= *VR41XX_MSYSINT1REG;
	irq_mask_probe[2] &= *VR41XX_MSYSINT2REG;
	irq_mask_probe[3] &= *VR41XX_MGIUINTLREG;
	irq_mask_probe[4] &= *VR41XX_MGIUINTHREG;

	return 0x12345678;
}

int probe_irq_off(unsigned long unused)
{
	int i, j, irq_found, nr_irqs;

	if (unused != 0x12345678)
		printk("Bad IRQ probe detected\n");

	irq_mask_probe[0] = read_32bit_cp0_register(CP0_STATUS) | ~irq_mask_probe[0];
	irq_mask_probe[1] = *VR41XX_MSYSINT1REG | ~irq_mask_probe[1];
	irq_mask_probe[2] = *VR41XX_MSYSINT2REG | ~irq_mask_probe[2];
	irq_mask_probe[3] = *VR41XX_MGIUINTLREG | ~irq_mask_probe[3];
	irq_mask_probe[4] = *VR41XX_MGIUINTHREG | ~irq_mask_probe[4];

	nr_irqs = 0;
	irq_found = 0;

	for (i = 0; i < 5; i++)
		for (j = 0; j < 16; j++)
			if (!(irq_mask_probe[i] & ((unsigned short)1 << j))) {
				irq_found = i * 16 + j - 8;
				nr_irqs++;
			}

	if (nr_irqs > 1)
		irq_found = -irq_found;
	return irq_found;
}

