/* $Id: tpanel.c,v 1.12 1999/11/09 22:14:24 mfklar Exp $
 *
 * Vr41xx Touch Panel Driver
 * by Michael Klar, wyldfier@iname.com
 *
 * Loosely based on drivers/char/busmouse.c
 */

// This is used to enable the /proc/piuinfo interface to return
// diagnostic info on PIU state and events:
// #define DEBUG_TIMING_PARAMS

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>

#ifdef DEBUG_TIMING_PARAMS
#include <linux/proc_fs.h>
#endif

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/vr41xx.h>

#include <linux/tpanel.h>

// BUFFSIZE can be increased, but it must be a power of 2

#define BUFFSIZE 128

struct tpanel_status {
	unsigned int	buffer[BUFFSIZE];
	unsigned char	head;
	unsigned char	tail;
	unsigned char	lostflag;
	wait_queue_head_t wait;
	struct fasync_struct *fasyncptr;
	int		active;
#ifdef DEBUG_TIMING_PARAMS
	unsigned short	orig_scanint;
	unsigned short	orig_settletime;
	unsigned int	dropped_count;
	unsigned int	overflow_count;
	unsigned int	incompletescan_count;
	unsigned int	dualpagevalid_count;
	unsigned int	unhandledint_count;
#endif
};

static struct tpanel_status tpanel;


static void tpanel_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int data1 = 0, data2 = 0;
	unsigned short intreg;

	while ((intreg = *VR41XX_PIUINTREG)) {

			// Panel contact state change interrupt
		if (intreg & VR41XX_PIUINTREG_PENCHG) {
			data1 = ((*VR41XX_PIUCNTREG & 0x2000) | 0x4000) << 17
					 | ((unsigned int)tpanel.lostflag << 28);
			tpanel.lostflag = 0;
			tpanel.buffer[tpanel.head++] = data1;
			tpanel.buffer[tpanel.head++] = 0;
			tpanel.head &= (BUFFSIZE - 1);
			if (tpanel.head == tpanel.tail) {
				tpanel.tail += 2;
				tpanel.tail &= (BUFFSIZE - 1);
				tpanel.buffer[tpanel.tail] |= 0x20000000;
#ifdef DEBUG_TIMING_PARAMS
				tpanel.dropped_count++;
#endif
			}

			*VR41XX_PIUINTREG |= VR41XX_PIUINTREG_PENCHG;
		}

			// PADCMD and PADADP shouldn't happen
		if (intreg & (VR41XX_PIUINTREG_PADCMD | VR41XX_PIUINTREG_PADADP)) {
			*VR41XX_PIUINTREG |= VR41XX_PIUINTREG_PADCMD |
					     VR41XX_PIUINTREG_PADADP;
#ifdef DEBUG_TIMING_PARAMS
			tpanel.unhandledint_count++;
#endif
		}

			// PADDLOST interrupt: datapoint lost, just report it
		if (intreg & VR41XX_PIUINTREG_PADDLOST) {
			tpanel.lostflag = 1;
#ifdef DEBUG_TIMING_PARAMS
			if ((intreg & (VR41XX_PIUINTREG_PADPAGE1 | VR41XX_PIUINTREG_PADPAGE0))
				 == (VR41XX_PIUINTREG_PADPAGE1 | VR41XX_PIUINTREG_PADPAGE0))
				tpanel.overflow_count++;
			else
				tpanel.incompletescan_count++;
#endif

			*VR41XX_PIUINTREG |= VR41XX_PIUINTREG_PADDLOST;
		}

			// Data buffer interrupt - NOTE: due to odd wording in
			// Vr41xx docs, the case where both pages are valid may
			// be handled in the wrong order.
		if (intreg & VR41XX_PIUINTREG_PADPAGE0) {
#ifdef DEBUG_TIMING_PARAMS
			if (intreg & VR41XX_PIUINTREG_PADPAGE1)
				tpanel.dualpagevalid_count++;
#endif
			if ((intreg & (VR41XX_PIUINTREG_PADPAGE1 | VR41XX_PIUINTREG_OVP))
				 != (VR41XX_PIUINTREG_PADPAGE1 | VR41XX_PIUINTREG_OVP)) {
				data1 = 0x40000000 |
					((unsigned int)tpanel.lostflag << 28) |
				        ((*VR41XX_PIUPB01REG & 0x3ff) << 10) |
					( *VR41XX_PIUPB00REG & 0x3ff);
				data2 = ( *VR41XX_PIUPB03REG << 20) |
					((*VR41XX_PIUPB02REG & 0x3ff) << 10) |
					( *VR41XX_PIUPB04REG & 0x3ff);
				tpanel.lostflag = 0;
				tpanel.buffer[tpanel.head++] = data1;
				tpanel.buffer[tpanel.head++] = data2;
				tpanel.head &= (BUFFSIZE - 1);
				if (tpanel.head == tpanel.tail) {
					tpanel.tail += 2;
					tpanel.tail &= (BUFFSIZE - 1);
					tpanel.buffer[tpanel.tail] |= 0x20000000;
#ifdef DEBUG_TIMING_PARAMS
					tpanel.dropped_count++;
#endif
				}

				*VR41XX_PIUINTREG |= VR41XX_PIUINTREG_PADPAGE0;
			}
		}
		if (intreg & VR41XX_PIUINTREG_PADPAGE1) {
			data1 = 0x40000000 |
				((unsigned int)tpanel.lostflag << 28) |
			        ((*VR41XX_PIUPB11REG & 0x3ff) << 10) |
				( *VR41XX_PIUPB10REG & 0x3ff);
			data2 = ( *VR41XX_PIUPB13REG << 20) |
				((*VR41XX_PIUPB12REG & 0x3ff) << 10) |
				( *VR41XX_PIUPB14REG & 0x3ff);
			tpanel.lostflag = 0;
			tpanel.buffer[tpanel.head++] = data1;
			tpanel.buffer[tpanel.head++] = data2;
			tpanel.head &= (BUFFSIZE - 1);
			if (tpanel.head == tpanel.tail) {
				tpanel.tail += 2;
				tpanel.tail &= (BUFFSIZE - 1);
				tpanel.buffer[tpanel.tail] |= 0x20000000;
#ifdef DEBUG_TIMING_PARAMS
				tpanel.dropped_count++;
#endif
			}

			*VR41XX_PIUINTREG |= VR41XX_PIUINTREG_PADPAGE1;
		}
	}
	if (data1) {
		add_mouse_randomness(data1);
		if (data2)
			add_mouse_randomness(data2);
		wake_up_interruptible(&tpanel.wait);

		if (tpanel.fasyncptr)
			kill_fasync(tpanel.fasyncptr, SIGIO);
	}

		// For some reason, bit PADSCANSTART gets set on occasion,
		// also PIUSEQEN will get reset if data is lost,  so we just
		// force CNTREG to a known state.  This hardware behavior
		// is apparently an undocumented "feature"
	barrier();
	*VR41XX_PIUCNTREG = 0x0326;
}

static int fasync_tpanel(int fd, struct file *filp, int on)
{
	int retval;

	retval = fasync_helper(fd, filp, on, &tpanel.fasyncptr);
	if (retval < 0)
		return retval;
	return 0;
}

static int close_tpanel(struct inode * inode, struct file * file)
{
	fasync_tpanel(-1, file, 0);
	if (--tpanel.active)
		return 0;
		// set for standby
	*VR41XX_MPIUINTREG = 0;
	*VR41XX_PIUCNTREG &= ~VR41XX_PIUCNTREG_PIUSEQEN;
	free_irq(VR41XX_IRQ_PIU, NULL);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int open_tpanel(struct inode * inode, struct file * file)
{
	if (tpanel.active++)
		return 0;
	*VR41XX_MPIUINTREG = 0;
	barrier();
	if (request_irq(VR41XX_IRQ_PIU, tpanel_interrupt, 0, "vr41xxtpanel", NULL)) {
		tpanel.active--;
		return -EBUSY;
	}
	tpanel.tail = tpanel.head;
	MOD_INC_USE_COUNT;
	barrier();

		// set up for autoscan
	*VR41XX_PIUCNTREG = 0x0322;	// autoscan, sequence suspended
	*VR41XX_PIUSIVLREG = 333;	// set interval to .01 sec default
	*VR41XX_PIUSTBLREG = 16;
	barrier();
	*VR41XX_MPIUINTREG = 0x007d;
	*VR41XX_PIUINTREG = 0x007d;	// clear any pending ints
	barrier();
	*VR41XX_PIUCNTREG = 0x0326;	// autoscan, sequence enabled
	return 0;
}

/*
 *  Read touch panel data.  Blocking read is not implemented, mostly to stay
 *  consistant with the way the mouse drivers do polling.
 *
 *  Data format:
 *  unsigned short status: bit 15 = xyz data valid (0 means contact state only)
 *                         bit 14 = pen contact state (1 means contact)
 *                         bit 13 = soft data lost flag: if this is 1, this point
 *                                  is valid (if bit 15 is 1), but data was lost
 *                                  between this point and the previous point
 *                                  due to a buffer overrun in the driver
 *                         bit 12 = hard data lost flag: if this is 1, this point
 *                                  is valid (if bit 15 is 1), but data was lost
 *                                  between this point and the previous point
 *                         bits 11-0 = reserved
 *  unsigned short x+ raw data (if status:15 = 1)
 *  unsigned short x- raw data (if status:15 = 1)
 *  unsigned short y+ raw data (if status:15 = 1)
 *  unsigned short y- raw data (if status:15 = 1)
 *  unsigned short z (pressure) raw data (if status:15 = 1)
 *
 *  x+, x-, y+, and y- are limited to range 0-1023 for this hardware.  Each +/-
 *  pair is somewhat redundant: the + value can be used as is, but using the
 *  difference (eg. (x+) - (x-)) will produce a more accurate result.  Further
 *  manipulation and/or statistical analysis may be required for best accuracy.
 *
 *  No calibration is done on the driver side, that is expected to be done on
 *  the user side.
 */

static ssize_t read_tpanel(struct file * file,
	char * buffer, size_t count, loff_t *ppos)
{
	int r;
	unsigned int data1, data2;
	unsigned long flags;

	// buffer too small?
	if (count < 12)
		return -EINVAL;

	// something wrong with the buffer address?
	if (!access_ok(VERIFY_WRITE, buffer, count))
		return -EFAULT;

		// We need to access the circular buffer atomic to
		// anything else that will read or write it
	save_flags(flags); cli();

	// no data available to return?
	if (tpanel.head == tpanel.tail) {
		restore_flags(flags);
		return -EAGAIN;
	}

		// Only return one packet of data, even if more are ready
	data1 = tpanel.buffer[tpanel.tail++];
	data2 = tpanel.buffer[tpanel.tail++];
	tpanel.tail &= BUFFSIZE - 1;

	restore_flags(flags);

	// put a single data point in the user buffer
	__put_user(((data1 >> 16) & 0xe000) ^ 0x8000,  (short*)buffer);
	__put_user((data1 >> 10) & 0x03ff, (short*)(buffer + 2));
	__put_user(data1 & 0x03ff,         (short*)(buffer + 4));
	__put_user((data2 >> 20) & 0x03ff, (short*)(buffer + 6));
	__put_user((data2 >> 10) & 0x03ff, (short*)(buffer + 8));
	__put_user(data2 & 0x03ff,         (short*)(buffer + 10));

	// clear any remaining user buffer area
	for (r=12 ; r < count; r++)
	    __put_user(0x00, buffer + r);

	return r;
}

static unsigned int poll_tpanel(struct file *file, poll_table * wait)
{
	poll_wait(file, &tpanel.wait, wait);
	if (tpanel.head != tpanel.tail)
		return POLLIN | POLLRDNORM;
	return 0;
}

static int ioctl_tpanel(struct inode * inode, struct file * file,
	unsigned int cmd, unsigned long arg)
{
	struct scanparam *sparm;
	unsigned int i1, i2;

	sparm = (struct scanparam *)arg;

	switch (cmd) {
	case TPGETSCANPARM:
		i1 = *VR41XX_PIUSIVLREG * 30;	// PIUSIVLREG is in 30us units
		if (put_user(i1, &sparm->interval))
			return -EFAULT;
		i2 = *VR41XX_PIUSTBLREG * 30;	// PIUSTBLREG is in 30us units
		if (put_user(i2, &sparm->settletime))
			return -EFAULT;
		return 0;
	case TPSETSCANPARM:
		if (get_user(i1, &sparm->interval))
			return -EFAULT;
		i1 = (i1 + 15) / 30;
		if (i1 > 0x07ff)
			return -EINVAL;
		if (get_user(i2, &sparm->settletime))
			return -EFAULT;
		i2 = (i2 + 15) / 30;
		if (i2 > 0x07ff)
			return -EINVAL;
		*VR41XX_PIUSIVLREG = (unsigned short)i1;
		*VR41XX_PIUSTBLREG = (unsigned short)i2;
		__put_user(i1 * 30, &sparm->interval);
		__put_user(i2 * 30, &sparm->settletime);
		return 0;
	default:
		return -EINVAL;
	}
}

struct file_operations vr41xx_tpanel_fops = {
	NULL,		/* seek */
	read_tpanel,
	NULL,		/* write */
	NULL, 		/* readdir */
	poll_tpanel,
	ioctl_tpanel,
	NULL,		/* mmap */
	open_tpanel,
	NULL,		/* flush */
	close_tpanel,
	NULL,
	fasync_tpanel,
};

static struct miscdevice vr41xx_tpanel = {
	VR41XX_TPANEL_MINOR, "vr41xxtpanel", &vr41xx_tpanel_fops
};

#ifdef DEBUG_TIMING_PARAMS
int get_piuinfo(char *buffer, char **buffer_location, off_t offset,
	int buffer_length, int zero)
{
	// return everything at once
	if ( offset > 0)
		return 0;

	// return lots of nifty diagnostic info
	return sprintf(buffer,
		"piucntreg:            0x%04hx\n"
		"piusivlreg:           0x%04hx\n"
		"piustblreg:           0x%04hx\n"
		"orig piusivlreg:      0x%04hx\n"
		"orig piustblreg:      0x%04hx\n"
		"dropped data count:   %d\n"
		"page overflow count:  %d\n"
		"bad scan count:       %d\n"
		"dual page count:      %d\n"
		"unhandled int count:  %d\n",
		*VR41XX_PIUCNTREG, *VR41XX_PIUSIVLREG, *VR41XX_PIUSTBLREG,
		tpanel.orig_scanint, tpanel.orig_settletime, tpanel.dropped_count,
		tpanel.overflow_count, tpanel.incompletescan_count,
		tpanel.dualpagevalid_count, tpanel.unhandledint_count);
}

struct proc_dir_entry piuinfo_entry = {
	0,                   // dynamic inode assigned by proc_register
	7, "piuinfo",        // name length and name
	S_IFREG | S_IRUGO,   // mode
	1,                   // number of links
	0, 0,                // uid, gid
	0,                   // size for ls
	NULL,                // inode ops
	get_piuinfo,         // get_info
	NULL                 // fill_inode
};
#endif

int __init vr41xx_tpanel_init(void)
{
	*VR41XX_CMUCLKMSK |= 0x0001;

#ifdef DEBUG_TIMING_PARAMS
	barrier();

	tpanel.orig_scanint = *VR41XX_PIUSIVLREG;
	tpanel.orig_settletime = *VR41XX_PIUSTBLREG;
	tpanel.dropped_count = 0;
	tpanel.overflow_count = 0;
	tpanel.incompletescan_count = 0;
	tpanel.dualpagevalid_count = 0;
	tpanel.unhandledint_count = 0;

	proc_register(&proc_root, &piuinfo_entry);
#endif

	*VR41XX_PIUCNTREG |= 0x0003;	// power on, but reset and suspend 
	
	tpanel.active = 0;
	tpanel.head = tpanel.tail = 0;
	init_waitqueue_head(&tpanel.wait);
	tpanel.fasyncptr = NULL;
	printk(KERN_INFO "Vr41xx touch panel initialized, using IRQ %d.\n",
	       VR41XX_IRQ_PIU);
	misc_register(&vr41xx_tpanel);
	return 0;
}

#ifdef MODULE

int init_module(void)
{
	return vr41xx_tpanel_init();
}

void cleanup_module(void)
{
	misc_deregister(&vr41xx_tpanel);
}
#endif

